#!
#gifar.pl
######################################################################
#
# Library for GIF arrangement
# Version 1.0
#
# Copyright 2000 (c) N.Saito
#
######################################################################
#
# (USE LIBRARY)
#
# To generate a GIF image for WWW counter, call the sub-routine
# 'generateImage' with two parametors, like this:
#
#   $result = &gifar'generateImage( 'srcimage.gif', '01234' );
#
# The return value is the packed binary of the generated image in
# GIF89a format.
# The first parametor is the filename and the path to the source image
# file that contains parts for the counter image.
# The second parametor is a string to generate image.Each character 
# represents dicimal letter, from '0' to '9'.
#
# (SOURCE IMAGE)
#
# Source image file is an animation GIF file that contains parts.
# Each frame of the animation is a image for a single number, and 
# sequenced in the order of the number from '0' to '9'. Transparent
# GIF also can be used.
#
# (TO GET NULL IMAGE)
#
# To get null image, just call like this:
#
#   $result = &gifar'getNullImage();
#
# This function returns a transparent image sized as 1x1 pixel.
#

package gifar;

$GIFHD = 'A6 S S C C C';
$GIFID = 'S S S S C';
$GIFGCE = 'C C C C S C C';
$TRANS = "0x47,0x49,0x46,0x38,0x39,0x61,0x01,0x00,0x01,0x00,0x80,0xFF,0x00,0xC0,0xC0,0xC0,0x00,0x00,0x00,0x21,0xF9,0x04,0x01,0x00,0x00,0x00,0x00,0x2C,0x00,0x00,0x00,0x00,0x01,0x00,0x01,0x00,0x00,0x02,0x02,0x44,0x01,0x00,0x3B";


# generateImage
########################################
# [Format]
#  $result = generateImage( $source_filename, $numstr );
sub generateImage {
 my( $file, $num ) = @_;
 my( $buf ) = &loadBinary( $file );
 my( %gh ) = &getHeader( \$buf );
 return '' unless ( $gh{'sig'} eq 'GIF89a' );

 my( @img ) = &separateImage( $buf );
 $buf = '';

 my( $ei, $lb, $bs, $f, $d, $t, $bt ) = unpack( $GIFGCE, $$src );

 my( $i, $n, $w );
 for ( $i=0, $w=0 ; $i<length($num) ; $i++ ) {
  $n = substr( $num, $i, 1 );
  $img[$n]{'gcef'} &= 0xe3;
  $img[$n]{'gcef'} |= ( 1 << 2 );    # No erase
  $buf .= pack( $GIFGCE, 0x21, 0xF9, 0x04, $img[$n]{'gcef'}, 
                         0, $img[$n]{'tci'}, 0 );
  $buf .= pack( 'C '.$GIFID, 0x2C, $w, 0, $img[$n]{'w'}, $img[$n]{'h'}, 
                             $img[$n]{'flags'} );
  $w += $img[$n]{'w'};
  $buf .= $img[$n]{'ct'} if ( $img[$n]{'ct'} );
  $buf .= $img[$n]{'image'};
 }
 $i = pack( $GIFHD, 'GIF89a', $w, $gh{'h'}, $gh{'flags'},
                    $gh{'bci'}, $gh{'par'} );
 $i .= $gh{'ct'} if ( $gh{'ct'} );
 return $i.$buf;
}

# getNullImage
########################################
# [Format]
#  $result = getNullImage();
sub getNullImage {
 return pack( 'C43', eval( "($TRANS)" ) );
}

# separateImage
########################################
sub separateImage {
 my( $buf ) = @_;
 my( $i, $k, $v, %h );
 my( @r ) = ();

 # Analyze data
 until ( ($i=unpack('C',$buf))==0x3B ) {    # Until find trailer
  if ( $i==0x2C ) {         # Image Desc
   %h = &getImageDesc( \$buf );
   $r[$#r]{$k}=$v while( ($k,$v)=each(%h) );
  } elsif ( $i==0x21 ) {    # Extensions
   if ( unpack('C',substr($buf,1))==0xF9 ) {    # Graphic Control
    push( @r, { &getGraphicControl(\$buf) } );
   } else {
    &getMisc( \$buf );
   }
  } else {
   return ();
  }
 }

 return @r;
}

# loadBinary
########################################
# [Format]
#  $result = loadBinary( $filename );
sub loadBinary {
 my( $file ) = @_;
 my( $size );
 my( $r ) = '';
 $size = -s $file;
 open( IN, $file ) || return $r;
 binmode( IN );
 sysread( IN, $r, $size );
 close( IN );
 return $r;
}

# getDataSubBlock
########################################
# [Format]
#  $result = getDataSubBlock( *src );
sub getDataSubBlock {
 my( $src ) = @_;
 my( $l, $s ) = ( 0, 0 );
 $s += $l+1 while ( ($l=unpack('C',substr($$src,$s))) );
 $s ++;
 $l = substr( $$src, 0, $s );
 $$src = substr( $$src, $s );
 return $l;
}

# getHeader
########################################
# [Format]
#  %result = getHeader( *src );
sub getHeader {
 my( $src ) = @_;
 my( @d ) = unpack( $GIFHD, $$src );
 my( %r ) = ();
 my( $i );
 $$src = substr( $$src, 13 );
 for ( $i=0 ; $i<@d ; $i++ ) {
  $r{ ('sig', 'w', 'h', 'flags', 'bci', 'par')[$i] } = $d[$i];
 }
 if ( $r{'flags'} & 0x80 ) {    # GCT is present
  $i = ( 1 << (($r{'flags'} & 0x07) + 1) ) * 3;
  $r{'ct'} = substr( $$src, 0, $i );
  $$src = substr( $$src, $i );
 }
 return %r;
}

# getImageDesc
########################################
# [Format]
#  %result = getImageDesc( *src );
sub getImageDesc {
 my( $src ) = @_;
 my( %r ) = ();
 return %r unless ( unpack('C',$$src)==0x2C );
# $r{'type'} = 0x2C;
 my( @d ) = unpack( $GIFID, substr($$src,1) );
 my( $i );
 $$src = substr( $$src, 10 );
 for ( $i=0 ; $i<@d ; $i++ ) {
  $r{ ('x','y','w','h','flags')[$i] } = $d[$i];
 }
 if ( $r{'flags'} & 0x80 ) {    # LCT is present
  $i = ( 1 << (($r{'flags'} & 0x07) + 1) ) * 3;
  $r{'ct'} = substr( $$src, 0, $i );
  $$src = substr( $$src, $i );
 }
 $i = substr( $$src, 0, 1 );
 $$src = substr( $$src, 1 );
 $r{'image'} = $i . getDataSubBlock( $src );
 return %r;
}

# getGraphicControl
########################################
# [Format]
#  %result = getGraphicControl( *src );
sub getGraphicControl {
 my( $src ) = @_;
 my( %r ) = ();
 my( $ei, $lb, $bs, $f, $d, $t, $bt ) = unpack( $GIFGCE, $$src );
 return %r unless ( ($ei==0x21) && ($lb==0xF9) && ($bs==0x04) && !$bt );
# $r{'type'} = 0x21;
# $r{'subtype'} = 0xF9;
 $r{'gcef'} = $f;
 $r{'delay'} = $d;
 $r{'tci'} = $t;
 $$src = substr( $$src, 8 );
 return %r;
}

# getMisc
########################################
# [Format]
#  %result = getMisc( *src );
sub getMisc {
 my( $src ) = @_;
 my( %r ) = ();
 return %r unless ( unpack('C',$$src)==0x21 );
 my( $st ) = unpack('C',substr($$src,1));
 return %r unless ( ($st==0x01) || ($st==0xFE) || ($st==0xFF) );
# $r{'type'} = 0x21;
# $r{'subtype'} = $st;
 $$src = substr( $$src, 2 );
 $r{'data'} = getDataSubBlock( $src );
 return %r;
}

